#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 A S Lewis
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.


"""Tidy operation classes."""


# Import Gtk modules
import gi
from gi.repository import GObject


# Import other modules
try:
    import moviepy.editor
except:
    pass

import os
import re
import shutil
import threading
import time


# Import our modules
import formats
import media
import utils
# Use same gettext translations
from mainapp import _


# Classes


class TidyManager(threading.Thread):

    """Called by mainapp.TartubeApp.tidy_manager_start().

    Python class to manage the tidy operation, in which videos can be checked
    for corruption and actually existing (or not), and various file types can
    be deleted collectively.

    Args:

        app_obj (mainapp.TartubeApp): The main application

        choices_dict (dict): A dictionary specifying the choices made by the
            user in mainwin.TidyDialogue. The dictionary is in the following
            format:

            media_data_obj: A media.Channel, media.Playlist or media.Folder
                object, or None if all channels/playlists/folders are to be
                tidied up. If specified, the cahnnel/playlist/folder and all of
                its descendants are checked

            corrupt_flag: True if video files should be checked for corruption

            del_corrupt_flag: True if corrupted video files should be deleted

            exist_Flag: True if video files that should exist should be
                checked, in case they don't (and vice-versa)

            del_video_flag: True if downloaded video files should be deleted

            del_others_flag: True if all video/audio files with the same name
                should be deleted (as artefacts of post-processing with FFmpeg
                or AVConv)

            del_archive_flag: True if all youtube-dl archive files should be
                deleted

            move_thumb_flag: True if all thumbnail files should be moved into a
                subdirectory

            del_thumb_flag: True if all thumbnail files should be deleted

            convert_webp_flag: True if all .webp thumbnail files should be
                converted to .jpg

            move_data_flag: True if description, metadata (JSON) and annotation
                files should be moved into a subdirectory

            del_descrip_flag: True if all description files should be deleted

            del_json_flag: True if all metadata (JSON) files should be deleted

            del_xml_flag: True if all annotation files should be deleted

    """


    # Standard class methods


    def __init__(self, app_obj, choices_dict):

        super(TidyManager, self).__init__()

        # IV list - class objects
        # -----------------------
        # The mainapp.TartubeApp object
        self.app_obj = app_obj
        # The media data object (channel, playlist or folder) to be tidied up,
        #   or None if the whole data directory is to be tidied up
        # If specified, the channel/playlist/folder and all of its descendants
        #   are checked
        self.init_obj = choices_dict['media_data_obj']


        # IV list - other
        # ---------------
        # Flag set to False if self.stop_tidy_operation() is called, which
        #   halts the operation immediately
        self.running_flag = True

        # The time at which the tidy operation began (in seconds since epoch)
        self.start_time = int(time.time())
        # The time at which the tidy operation completed (in seconds since
        #   epoch)
        self.stop_time = None
        # The time (in seconds) between iterations of the loop in self.run()
        self.sleep_time = 0.25

        # Flags specifying which actions should be applied
        # True if video files should be checked for corruption
        self.corrupt_flag = choices_dict['corrupt_flag']
        # True if corrupted video files should be deleted
        self.del_corrupt_flag = choices_dict['del_corrupt_flag']
        # True if video files that should exist should be checked, in case they
        #   don't (and vice-versa)
        self.exist_flag = choices_dict['exist_flag']
        # True if downloaded video files should be deleted
        self.del_video_flag = choices_dict['del_video_flag']
        # True if all video/audio files with the same name should be deleted
        #   (as artefacts of post-processing with FFmpeg or AVConv)
        self.del_others_flag = choices_dict['del_others_flag']
        # True if all youtube-dl archive files should be deleted
        self.del_archive_flag = choices_dict['del_archive_flag']
        # True if all thumbnail files should be moved into a subdirectory
        self.move_thumb_flag = choices_dict['move_thumb_flag']
        # True if all thumbnail files should be deleted
        self.del_thumb_flag = choices_dict['del_thumb_flag']
        # True if all .webp thumbnail files should be converted to .jpg.
        #   Requires mainapp.TartubeApp.ffmpeg_fail_flag set to False
        self.convert_webp_flag = choices_dict['convert_webp_flag']
        # True if description, metadata (JSON) and annotation files should be
        #   moved into a subdirectory
        self.move_data_flag = choices_dict['move_data_flag']
        # True if all description files should be deleted
        self.del_descrip_flag = choices_dict['del_descrip_flag']
        # True if all metadata (JSON) files should be deleted
        self.del_json_flag = choices_dict['del_json_flag']
        # True if all annotation files should be deleted
        self.del_xml_flag = choices_dict['del_xml_flag']

        # The number of media data objects whose directories have been tidied
        #   so far...
        self.job_count = 0
        # ...and the total number to tidy (these numbers are displayed in the
        #   progress bar in the Videos tab)
        self.job_total = 0

        # Individual counts, updated as we go
        self.video_corrupt_count = 0
        self.video_corrupt_deleted_count = 0
        self.video_exist_count = 0
        self.video_no_exist_count = 0
        self.video_deleted_count = 0
        self.other_deleted_count = 0
        self.archive_deleted_count = 0
        self.thumb_moved_count = 0
        self.thumb_deleted_count = 0
        self.webp_converted_count = 0
        self.data_moved_count = 0
        self.descrip_deleted_count = 0
        self.json_deleted_count = 0
        self.xml_deleted_count = 0


        # Code
        # ----

        # Do not convert .webp thumbnails, if not allowed
        if self.app_obj.ffmpeg_fail_flag:
            self.convert_webp_flag = False

        # Let's get this party started!
        self.start()


    # Public class methods


    def run(self):

        """Called as a result of self.__init__().

        Compiles a list of media data objects (channels, playlists and folders)
        to tidy up. If self.init_obj is not set, only that channel/playlist/
        folder (and its child channels/playlists/folders) are tidied up;
        otherwise the whole data directory is tidied up.

        Then calls self.tidy_directory() for each item in the list.

        Finally informs the main application that the tidy operation is
        complete.
        """

        # Show information about the tidy operation in the Output tab
        if not self.init_obj:
            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                _('Starting tidy operation, tidying up whole data directory'),
            )

        else:

            media_type = self.init_obj.get_type()

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                _('Starting tidy operation, tidying up \'{0}\'').format(
                    self.init_obj.name,
                )
            )

        if self.corrupt_flag:
            text = _('YES')
        else:
            text = _('NO')

        self.app_obj.main_win_obj.output_tab_write_stdout(
            1,
            '   ' + _('Check videos are not corrupted:') + ' ' + text,
        )

        if self.corrupt_flag:

            if self.del_corrupt_flag:
                text = _('YES')
            else:
                text = _('NO')

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('Delete corrupted videos:') + ' ' + text,
            )

        if self.exist_flag:
            text = _('YES')
        else:
            text = _('NO')

        self.app_obj.main_win_obj.output_tab_write_stdout(
            1,
            '   ' + _('Check videos do/don\'t exist:') + ' ' + text,
        )

        if self.del_video_flag:
            text = _('YES')
        else:
            text = _('NO')

        self.app_obj.main_win_obj.output_tab_write_stdout(
            1,
            '   ' + _('Delete all video files:') + ' ' + text,
        )

        if self.del_video_flag:

            if self.del_others_flag:
                text = _('YES')
            else:
                text = _('NO')

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('Delete other video/audio files:') + ' ' + text,
            )

        if self.del_archive_flag:
            text = _('YES')
        else:
            text = _('NO')

        self.app_obj.main_win_obj.output_tab_write_stdout(
            1,
            '   ' + _('Delete downloader archive files:') + ' ' + text,
        )

        if self.move_thumb_flag:
            text = _('YES')
        else:
            text = _('NO')

        self.app_obj.main_win_obj.output_tab_write_stdout(
            1,
            '   ' + _('Move thumbnails into own folder:') + ' ' + text,
        )

        if self.del_thumb_flag:
            text = _('YES')
        else:
            text = _('NO')

        self.app_obj.main_win_obj.output_tab_write_stdout(
            1,
            '   ' + _('Delete all thumbnail files:') + ' ' + text,
        )

        if self.convert_webp_flag:
            text = _('YES')
        else:
            text = _('NO')

        self.app_obj.main_win_obj.output_tab_write_stdout(
            1,
            '   ' + _('Convert .webp thumbnails to .jpg:') + ' ' + text,
        )

        if self.move_data_flag:
            text = _('YES')
        else:
            text = _('NO')

        self.app_obj.main_win_obj.output_tab_write_stdout(
            1,
            '   ' + _('Move other metadata files into own folder:') \
            + ' ' + text,
        )

        if self.del_descrip_flag:
            text = _('YES')
        else:
            text = _('NO')

        self.app_obj.main_win_obj.output_tab_write_stdout(
            1,
            '   ' + _('Delete all description files:') + ' ' + text,
        )

        if self.del_json_flag:
            text = _('YES')
        else:
            text = _('NO')

        self.app_obj.main_win_obj.output_tab_write_stdout(
            1,
            '   ' + _('Delete all metadata (JSON) files:') + ' ' + text,
        )

        if self.del_xml_flag:
            text = _('YES')
        else:
            text = _('NO')

        self.app_obj.main_win_obj.output_tab_write_stdout(
            1,
            '   ' + _('Delete all annotation files:') + ' ' + text,
        )

        # Compile a list of channels, playlists and folders to tidy up (each
        #   one has their own sub-directory inside Tartube's data directory)
        obj_list = []
        if self.init_obj:
            # Add this channel/playlist/folder, and any child channels/
            #   playlists/folders (but not videos, obviously)
            obj_list = self.init_obj.compile_all_containers(obj_list)
        else:
            # Add all channels/playlists/folders in the database
            for dbid in list(self.app_obj.media_name_dict.values()):

                obj = self.app_obj.media_reg_dict[dbid]
                # Don't add private folders
                if not isinstance(obj, media.Folder) or not obj.priv_flag:
                    obj_list.append(obj)

        self.job_total = len(obj_list)

        # Check each sub-directory in turn, updating the media data registry
        #   as we go
        while self.running_flag and obj_list:
            self.tidy_directory(obj_list.pop(0))

            # Pause a moment, before the next iteration of the loop (don't want
            #   to hog resources)
            time.sleep(self.sleep_time)

        # Operation complete. Set the stop time
        self.stop_time = int(time.time())

        # Show a confirmation in the Output tab
        self.app_obj.main_win_obj.output_tab_write_stdout(
            1,
            _('Tidy operation finished'),
        )

        if self.corrupt_flag:

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('Corrupted videos found:') + ' ' \
                + str(self.video_corrupt_count),
            )

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('Corrupted videos deleted:') + ' ' \
                + str(self.video_corrupt_deleted_count),
            )

        if self.exist_flag:

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('New video files detected:') + ' ' \
                + str(self.video_exist_count),
            )

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('Missing video files detected:') + ' ' \
                + str(self.video_no_exist_count),
            )

        if self.del_video_flag:

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('Non-corrupted video files deleted:') + ' ' \
                + str(self.video_deleted_count),
            )

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('Other video/audio files deleted:') + ' ' \
                + str(self.other_deleted_count),
            )

        if self.del_archive_flag:

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('Downloader archive files deleted:') + ' ' \
                + str(self.archive_deleted_count),
            )

        if self.move_thumb_flag:

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('Thumbnail files moved:') + ' ' \
                + str(self.thumb_moved_count),
            )

        if self.del_thumb_flag:

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('Thumbnail files deleted:') + ' ' \
                + str(self.thumb_deleted_count),
            )

        if self.convert_webp_flag:

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('.webp thumbnails converted to .jpg:') + ' ' \
                + str(self.webp_converted_count),
            )

        if self.move_data_flag:

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('Other metadata files moved:') + ' ' \
                + str(self.data_moved_count),
            )

        if self.del_descrip_flag:

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('Description files deleted:') + ' ' \
                + str(self.descrip_deleted_count),
            )

        if self.del_json_flag:

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('Metadata (JSON) files deleted:') + ' ' \
                + str(self.json_deleted_count),
            )

        if self.del_xml_flag:

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('Annotation files deleted:') + ' ' \
                + str(self.xml_deleted_count),
            )

        # Let the timer run for a few more seconds to prevent Gtk errors (for
        #   systems with Gtk < 3.24)
        GObject.timeout_add(
            0,
            self.app_obj.tidy_manager_halt_timer,
        )


    def tidy_directory(self, media_data_obj):

        """Called by self.run().

        Tidy up the directory of a single channel, playlist or folder.

        Args:

            media_data_obj (media.Channel, media.Playlist or media.Folder):
                The media data object whose directory must be tidied up

        """

        # Update the main window's progress bar
        self.job_count += 1
        GObject.timeout_add(
            0,
            self.app_obj.main_win_obj.update_progress_bar,
            media_data_obj.name,
            self.job_count,
            self.job_total,
        )

        media_type = media_data_obj.get_type()

        self.app_obj.main_win_obj.output_tab_write_stdout(
            1,
            _('Checking:') + ' \'' + media_data_obj.name + '\'',
        )

        if self.corrupt_flag:
            self.check_video_corrupt(media_data_obj)

        if self.exist_flag:
            self.check_videos_exist(media_data_obj)

        if self.del_video_flag:
            self.delete_video(media_data_obj)

        if self.del_archive_flag:
            self.delete_archive(media_data_obj)

        if self.move_thumb_flag:
            self.move_thumb(media_data_obj)

        if self.del_thumb_flag:
            self.delete_thumb(media_data_obj)

        if self.convert_webp_flag:
            self.convert_webp(media_data_obj)

        if self.move_data_flag:
            self.move_data(media_data_obj)

        if self.del_descrip_flag:
            self.delete_descrip(media_data_obj)

        if self.del_json_flag:
            self.delete_json(media_data_obj)

        if self.del_xml_flag:
            self.delete_xml(media_data_obj)


    def check_video_corrupt(self, media_data_obj):

        """Called by self.tidy_directory().

        Checks all child videos of the specified media data object. If the
        video are corrupted, don't delete them (let the user do that manually).

        Args:

            media_data_obj (media.Channel, media.Playlist or media.Folder):
                The media data object whose directory must be tidied up

        """

        # Import the main window (for convenience)
        main_win_obj = self.app_obj.main_win_obj

        for video_obj in media_data_obj.compile_all_videos( [] ):

            if video_obj.file_name is not None \
            and video_obj.dl_flag:

                video_path = video_obj.get_actual_path(self.app_obj)

                if os.path.isfile(video_path):

                    # Code copied from
                    #   mainapp.TartubeApp.update_video_from_filesystem()
                    # When the video file is corrupted, moviepy freezes
                    #   indefinitely
                    # Instead, let's try placing the procedure inside a thread
                    #   (unlike the original function, this one is never called
                    #   if .refresh_moviepy_timeout is 0)
                    this_thread = threading.Thread(
                        target=self.call_moviepy,
                        args=(video_obj, video_path,),
                    )

                    this_thread.daemon = True
                    this_thread.start()
                    this_thread.join(self.app_obj.refresh_moviepy_timeout)
                    if this_thread.is_alive():

                        # moviepy timed out, so assume the video is corrupted
                        self.video_corrupt_count += 1

                        if self.del_corrupt_flag \
                        and os.path.isfile(video_path):

                            # Delete the corrupted file
                            if self.app_obj.remove_file(video_path):
                                self.video_corrupt_deleted_count += 1

                                main_win_obj.output_tab_write_stdout(
                                    1,
                                    '   ' + _(
                                    'Deleted (possibly) corrupted video file:',
                                    ) + ' \'' + video_obj.name + '\'',
                                )

                                self.app_obj.mark_video_downloaded(
                                    video_obj,
                                    False,
                                )

                            else:
                                main_win_obj.output_tab_write_stderr(
                                    1,
                                    '   ' + _(
                                        'Failed to delete (possibly)' \
                                        + ' corrupted video file:',
                                    ) + ' \'' + video_obj.name + '\'',
                                )

                        else:

                            # Don't delete it
                            self.app_obj.main_win_obj.output_tab_write_stdout(
                                1,
                                '   ' + _(
                                    'Video file might be corrupt:',
                                ) + ' \'' + video_obj.name + '\'',
                            )


    def check_videos_exist(self, media_data_obj):

        """Called by self.tidy_directory().

        Checks all child videos of the specified media data object. If the
        video should exist, but doesn't (or vice-versa), modify the media.Video
        object's IVs accordingly.

        Args:

            media_data_obj (media.Channel, media.Playlist or media.Folder):
                The media data object whose directory must be tidied up

        """

        for video_obj in media_data_obj.compile_all_videos( [] ):

            if video_obj.file_name is not None:

                video_path = video_obj.get_actual_path(self.app_obj)

                if not video_obj.dl_flag \
                and os.path.isfile(video_path):

                    # File exists, but is marked as not downloaded
                    self.app_obj.mark_video_downloaded(
                        video_obj,
                        True,       # Video is downloaded
                        True,       # ...but don't mark it as new
                    )

                    self.video_exist_count += 1

                    self.app_obj.main_win_obj.output_tab_write_stdout(
                        1,
                        '   ' + _(
                            'Video file exists:',
                        ) + ' \'' + video_obj.name + '\'',
                    )

                elif video_obj.dl_flag \
                and not os.path.isfile(video_path):

                    # File doesn't exist, but is marked as downloaded
                    self.app_obj.mark_video_downloaded(
                        video_obj,
                        False,      # Video is not downloaded
                    )

                    self.video_no_exist_count += 1

                    self.app_obj.main_win_obj.output_tab_write_stdout(
                        1,
                        '   ' + _(
                            'Video file doesn\'t exist:',
                        ) + ' \'' + video_obj.name + '\'',
                    )


    def delete_video(self, media_data_obj):

        """Called by self.tidy_directory().

        Checks all child videos of the specified media data object. If the
        video exists, delete it.

        Args:

            media_data_obj (media.Channel, media.Playlist or media.Folder):
                The media data object whose directory must be tidied up

        """

        ext_list = formats.VIDEO_FORMAT_LIST.copy()
        ext_list.extend(formats.AUDIO_FORMAT_LIST)

        for video_obj in media_data_obj.compile_all_videos( [] ):

            video_path = None
            if video_obj.file_name is not None:

                video_path = video_obj.get_actual_path(self.app_obj)

                # If the video's parent container has an alternative download
                #   destination set, we must check the corresponding media
                #   data object. If the latter also has a media.Video object
                #   matching this video, then this function returns None and
                #   nothing is deleted
                video_path = self.check_video_in_actual_dir(
                    media_data_obj,
                    video_obj,
                    video_path,
                )

            if video_path is not None:

                if video_obj.dl_flag \
                and os.path.isfile(video_path):

                    # Delete the downloaded video file
                    if self.app_obj.remove_file(video_path):

                        # Mark the video as not downloaded
                        self.app_obj.mark_video_downloaded(video_obj, False)
                        self.video_deleted_count += 1

                if self.del_others_flag:

                    # Also delete all video/audio files with the same name
                    # There might be thousands of files in the directory, so
                    #   using os.walk() or something like that might be too
                    #   expensive
                    # Also, post-processing might create various artefacts, all
                    #   of which must be deleted
                    for ext in ext_list:

                        other_path = video_obj.get_actual_path_by_ext(
                            self.app_obj,
                            ext,
                        )

                        if os.path.isfile(other_path) \
                        and self.app_obj.remove_file(other_path):
                            self.other_deleted_count += 1

        # For an encore, delete all post-processing artefacts in the form
        #   VIDEO_NAME.fNNN.ext, where NNN is an integer and .ext is one of
        #   the video extensions specified by formats.VIDEO_FORMAT_LIST
        #   (.mkv, etc)
        # (The previous code won't pick them up, but we can delete them all
        #   now.)
        # (The alternative download destination, if set, is not affected.)
        check_list = []
        search_path = media_data_obj.get_default_dir(self.app_obj)

        for (dir_path, dir_name_list, file_name_list) in os.walk(search_path):
            check_list.extend(file_name_list)

        char = '|'
        regex = '\.f\d+\.(' + char.join(formats.VIDEO_FORMAT_LIST) + ')$'
        for check_path in check_list:
            if re.search(regex, check_path):

                full_path = os.path.abspath(
                    os.path.join(search_path, check_path),
                )

                if os.path.isfile(full_path) \
                and self.app_obj.remove_file(full_path):
                    self.other_deleted_count += 1


    def delete_archive(self, media_data_obj):

        """Called by self.tidy_directory().

        Checks the specified media data object's directory. If a youtube-dl
        archive file is found there, delete it.

        Args:

            media_data_obj (media.Channel, media.Playlist or media.Folder):
                The media data object whose directory must be tidied up

        """

        archive_path = os.path.abspath(
            os.path.join(
                media_data_obj.get_default_dir(self.app_obj),
                'ytdl-archive.txt',
            ),
        )

        # Delete the archive file
        if os.path.isfile(archive_path) \
        and self.app_obj.remove_file(archive_path):
            self.archive_deleted_count += 1


    def move_thumb(self, media_data_obj):

        """Called by self.tidy_directory().

        Checks all child videos of the specified media data object. If the
        associated thumbnail file exists, moves it into its own sub-directory.

        Args:

            media_data_obj (media.Channel, media.Playlist or media.Folder):
                The media data object whose directory must be tidied up

        """

        for video_obj in media_data_obj.compile_all_videos( [] ):

            if video_obj.file_name is not None:

                # Thumbnails might be in one of four locations. If the
                #   thumbnail has already been moved into /.thumbs, then of
                #   course we don't move it again (and this function returns an
                #   empty list)
                path_list = utils.find_thumbnail_restricted(
                    self.app_obj,
                    video_obj,
                )

                if path_list:

                    main_path = os.path.abspath(
                        os.path.join(
                            path_list[0], path_list[1],
                        ),
                    )

                    subdir = os.path.abspath(
                        os.path.join(
                            path_list[0], self.app_obj.thumbs_sub_dir,
                        ),
                    )

                    subdir_path = os.path.abspath(
                        os.path.join(
                            path_list[0],
                            self.app_obj.thumbs_sub_dir,
                            path_list[1],
                        ),
                    )

                    if os.path.isfile(main_path) \
                    and not os.path.isfile(subdir_path):

                        if not os.path.isdir(subdir):
                            self.app_obj.make_directory(subdir)

                        if self.app_obj.move_file_or_directory(
                            main_path,
                            subdir_path,
                        ):
                            self.thumb_moved_count += 1


    def delete_thumb(self, media_data_obj):

        """Called by self.tidy_directory().

        Checks all child videos of the specified media data object. If the
        associated thumbnail file exists, delete it.

        Args:

            media_data_obj (media.Channel, media.Playlist or media.Folder):
                The media data object whose directory must be tidied up

        """

        for video_obj in media_data_obj.compile_all_videos( [] ):

            if video_obj.file_name is not None:

                # Thumbnails might be in one of four locations
                thumb_path = utils.find_thumbnail(self.app_obj, video_obj)

                # If the video's parent container has an alternative download
                #   destination set, we must check the corresponding media
                #   data object. If the latter also has a media.Video object
                #   matching this video, then this function returns None and
                #   nothing is deleted
                if thumb_path is not None:

                    thumb_path = self.check_video_in_actual_dir(
                        media_data_obj,
                        video_obj,
                        thumb_path,
                    )

                # Delete the thumbnail file
                if thumb_path is not None \
                and os.path.isfile(thumb_path) \
                and self.app_obj.remove_file(thumb_path):
                    self.thumb_deleted_count += 1


    def convert_webp(self, media_data_obj):

        """Called by self.tidy_directory().

        Checks all child videos of the specified media data object. If the
        associated thumbnail file in a .webp or malformed .jpg format exists,
        convert it to .jpg.

        Args:

            media_data_obj (media.Channel, media.Playlist or media.Folder):
                The media data object whose directory must be tidied up

        """

        for video_obj in media_data_obj.compile_all_videos( [] ):

            if video_obj.file_name is not None:

                # Thumbnails might be in one of four locations
                thumb_path = utils.find_thumbnail_webp(self.app_obj, video_obj)

                # If the video's parent container has an alternative download
                #   destination set, we must check the corresponding media
                #   data object. If the latter also has a media.Video object
                #   matching this video, then this function returns None and
                #   nothing is deleted
                if thumb_path is not None:

                    thumb_path = self.check_video_in_actual_dir(
                        media_data_obj,
                        video_obj,
                        thumb_path,
                    )

                if thumb_path is not None \
                and os.path.isfile(thumb_path):

                    # Convert to .jpg
                    if not self.app_obj.ffmpeg_manager_obj.convert_webp(
                        thumb_path
                    ):
                        # FFmpeg is probably not installed; don't try any more
                        #   conversions
                        self.convert_webp_flag = False
                        self.app_obj.set_ffmpeg_fail_flag(True)

                    else:

                        self.webp_converted_count += 1


    def move_data(self, media_data_obj):

        """Called by self.tidy_directory().

        Checks all child videos of the specified media data object. If the
        associated thumbnail file exists, moves it into its own sub-directory.

        Args:

            media_data_obj (media.Channel, media.Playlist or media.Folder):
                The media data object whose directory must be tidied up

        """

        for video_obj in media_data_obj.compile_all_videos( [] ):

            if video_obj.file_name is not None:

                # Description/JSON/annotations files might be in one of four
                #   locations. If the file has already been moved into /.data,
                #   then of course we don't move it again
                for ext in ['.description', '.info.json', '.annotations.xml']:

                    main_path = video_obj.get_actual_path_by_ext(
                        self.app_obj,
                        ext,
                    )

                    subdir = os.path.abspath(
                        os.path.join(
                            video_obj.parent_obj.get_actual_dir(self.app_obj),
                            self.app_obj.metadata_sub_dir,
                        ),
                    )

                    subdir_path \
                    = video_obj.get_actual_path_in_subdirectory_by_ext(
                        self.app_obj,
                        ext,
                    )

                    if os.path.isfile(main_path) \
                    and not os.path.isfile(subdir_path):

                        if not os.path.isdir(subdir):
                            self.app_obj.make_directory(subdir)

                        # (os.rename sometimes fails on external hard drives;
                        #   this is safer)
                        if self.app_obj.move_file_or_directory(
                            main_path,
                            subdir_path,
                        ):
                            self.data_moved_count += 1


    def delete_descrip(self, media_data_obj):

        """Called by self.tidy_directory().

        Checks all child videos of the specified media data object. If the
        associated description file exists, delete it.

        Args:

            media_data_obj (media.Channel, media.Playlist or media.Folder):
                The media data object whose directory must be tidied up

        """

        for video_obj in media_data_obj.compile_all_videos( [] ):

            if video_obj.file_name is not None:

                main_path = video_obj.get_actual_path_by_ext(
                    self.app_obj,
                    '.description',
                )

                # If the video's parent container has an alternative download
                #   destination set, we must check the corresponding media
                #   data object. If the latter also has a media.Video object
                #   matching this video, then this function returns None and
                #   nothing is deleted
                main_path = self.check_video_in_actual_dir(
                    media_data_obj,
                    video_obj,
                    main_path,
                )

                # Delete the description file
                if main_path is not None \
                and os.path.isfile(main_path) \
                and self.app_obj.remove_file(main_path):
                    self.descrip_deleted_count += 1

                # (Repeat for a file that might be in the sub-directory
                #   '.data')
                subdir_path = video_obj.get_actual_path_in_subdirectory_by_ext(
                    self.app_obj,
                    '.description',
                )

                subdir_path = self.check_video_in_actual_dir(
                    subdir_path,
                    video_obj,
                    subdir_path,
                )

                if subdir_path is not None \
                and os.path.isfile(subdir_path) \
                and self.app_obj.remove_file(subdir_path):
                    self.descrip_deleted_count += 1


    def delete_json(self, media_data_obj):

        """Called by self.tidy_directory().

        Checks all child videos of the specified media data object. If the
        associated metadata (JSON) file exists, delete it.

        Args:

            media_data_obj (media.Channel, media.Playlist or media.Folder):
                The media data object whose directory must be tidied up

        """

        for video_obj in media_data_obj.compile_all_videos( [] ):

            if video_obj.file_name is not None:

                main_path = video_obj.get_actual_path_by_ext(
                    self.app_obj,
                    '.info.json',
                )

                # If the video's parent container has an alternative download
                #   destination set, we must check the corresponding media
                #   data object. If the latter also has a media.Video object
                #   matching this video, then this function returns None and
                #   nothing is deleted
                main_path = self.check_video_in_actual_dir(
                    media_data_obj,
                    video_obj,
                    main_path,
                )

                # Delete the metadata file
                if main_path is not None \
                and os.path.isfile(main_path) \
                and self.app_obj.remove_file(main_path):
                    self.json_deleted_count += 1

                # (Repeat for a file that might be in the sub-directory
                #   '.data')
                subdir_path = video_obj.get_actual_path_in_subdirectory_by_ext(
                    self.app_obj,
                    '.info.json',
                )

                subdir_path = self.check_video_in_actual_dir(
                    media_data_obj,
                    video_obj,
                    subdir_path,
                )

                if subdir_path is not None \
                and os.path.isfile(subdir_path) \
                and self.app_obj.remove_file(subdir_path):
                    self.json_deleted_count += 1


    def delete_xml(self, media_data_obj):

        """Called by self.tidy_directory().

        Checks all child videos of the specified media data object. If the
        associated annotation file exists, delete it.

        Args:

            media_data_obj (media.Channel, media.Playlist or media.Folder):
                The media data object whose directory must be tidied up

        """

        for video_obj in media_data_obj.compile_all_videos( [] ):

            if video_obj.file_name is not None:

                main_path = video_obj.get_actual_path_by_ext(
                    self.app_obj,
                    '.annotations.xml',
                )

                # If the video's parent container has an alternative download
                #   destination set, we must check the corresponding media
                #   data object. If the latter also has a media.Video object
                #   matching this video, then this function returns None and
                #   nothing is deleted
                main_path = self.check_video_in_actual_dir(
                    media_data_obj,
                    video_obj,
                    main_path,
                )

                # Delete the annotation file
                if main_path is not None \
                and os.path.isfile(main_path) \
                and self.app_obj.remove_file(main_path):
                    self.xml_deleted_count += 1

                # (Repeat for a file that might be in the sub-directory
                #   '.data')
                subdir_path = video_obj.get_actual_path_in_subdirectory_by_ext(
                    self.app_obj,
                    '.annotations.xml',
                )

                subdir_path = self.check_video_in_actual_dir(
                    media_data_obj,
                    video_obj,
                    subdir_path,
                )

                if subdir_path is not None \
                and os.path.isfile(subdir_path) \
                and self.app_obj.remove_file(subdir_path):
                    self.xml_deleted_count += 1


    def call_moviepy(self, video_obj, video_path):

        """Called by thread inside self.check_video_corrupt().

        When we call moviepy.editor.VideoFileClip() on a corrupted video file,
        moviepy freezes indefinitely.

        This function is called inside a thread, so a timeout of (by default)
        ten seconds can be applied.

        Args:

            video_obj (media.Video): The video object being updated

            video_path (str): The path to the video file itself

        """

        try:
            clip = moviepy.editor.VideoFileClip(video_path)

        except:
            self.video_corrupt_count += 1

            self.app_obj.main_win_obj.output_tab_write_stdout(
                1,
                '   ' + _('Video file might be corrupt:') + ' \'' \
                + video_obj.name + '\'',
            )


    def check_video_in_actual_dir(self, container_obj, video_obj, delete_path):

        """Called by self.delete_video(), .delete_descrip(), .delete_json(),
        .delete_xml() and .delete_thumb().

        If the video's parent container has an alternative download destination
        set, we must check the corresponding media data object. If the latter
        also has a media.Video object matching this video, then this function
        returns None and nothing is deleted. Otherwise, the specified
        delete_path is returned, so it can be deleted.

        Args:

            container_obj (media.Channel, media.Playlist, media.Folder): A
                channel, playlist or folder

            video_obj (media.Video): A video contained in that channel,
                playlist or folder

            delete_path (str): The path to a file which the calling function
                wants to delete

        Returns:

            The specified delete_path if it can be deleted, or None if it
                should not be deleted

        """

        if container_obj.external_dir is not None \
        or container_obj.dbid == container_obj.master_dbid:

            # No alternative download destination to check
            return delete_path

        else:

            # Get the channel/playlist/folder acting as container_obj's
            #   alternative download destination
            master_obj = self.app_obj.media_reg_dict[container_obj.master_dbid]

            # Check its videos. Are there any videos with the same name?
            for child_obj in master_obj.child_list:

                if child_obj.file_name is not None \
                and child_obj.file_name == video_obj.file_name:

                    # Don't delete the file associated with this video
                    return None

            # There are no videos with the same name, so the file can be
            #   deleted
            return delete_path


    def stop_tidy_operation(self):

        """Called by mainapp.TartubeApp.do_shutdown(), .stop_continue(),
        .on_button_stop_operation() and mainwin.MainWin.on_stop_menu_item().

        Stops the tidy operation.
        """

        self.running_flag = False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 A S Lewis
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.


"""FFmpeg manager classes."""


# Import Gtk modules
#   ...


# Import other modules
import os
import re
import subprocess


# Import our modules
import mainapp
import utils


# Classes


class FFmpegManager(object):

    """Called by mainapp.TartubeApp.__init__().

    Python class to manage calls to FFmpeg that Tartube wants to make,
    independently of youtube-dl.

    Most of the code in this file has been updated from youtube-dl itself.

    Args:

        app_obj (mainapp.TartubeApp): The main application object

    """


    # Standard class methods


    def __init__(self, app_obj):


        super(FFmpegManager, self).__init__()

        # IV list - class objects
        # -----------------------
        # The main application
        self.app_obj = app_obj


    # Public class methods


    def convert_webp(self, thumbnail_filename):

        """Called by mainapp.TartubeApp.update_video_when_file_found(),
        downloads.VideoDownloader.confirm_sim_video() and
        downloads.JSONFetcher.do_fetch().

        Adapted from youtube-dl/youtube-dl/postprocessor/embedthumbnail.py.

        In June 2020, YouTube changed its image format from .jpg to .webp.
        Unfortunately, the Gtk library doesn't support that format.

        Worse still, YouTube also began sending .webp thumbnails mislabelled as
        .jpg.

        In response, in September 2020 youtube-dl implemented a fix for
        embedded thumbnails, using FFmpeg to convert .webp to .jpg. That code
        has been adapted here, so that YouTube thumbnails can be converted and
        made visible in the main window again.

        Args:

            thumbnail_filename (str): Full path to the webp file to be
                converted to jpg

        Returns:

            False if an attempted conversion fails, or True otherwise
                (including when no conversion is attempted)

        """

        # Sanity check
        if not os.path.isfile(thumbnail_filename) \
        or self.app_obj.ffmpeg_fail_flag:
            return True

        # Correct extension for .webp files with the wrong extension
        #   (youtube-dl #25687, #25717)
        _, thumbnail_ext = os.path.splitext(thumbnail_filename)
        if thumbnail_ext:

            # Remove the initial full stop
            thumbnail_ext = thumbnail_ext[1:].lower()

            if thumbnail_ext != 'webp' and self.is_webp(thumbnail_filename):

                # .webp mislabelled as .jpg
                thumbnail_webp_filename = self.replace_extension(
                    thumbnail_filename,
                    'webp',
                )

                os.rename(thumbnail_filename, thumbnail_webp_filename)
                thumbnail_filename = thumbnail_webp_filename
                thumbnail_ext = 'webp'

        # Convert unsupported thumbnail formats to JPEG
        #   (youtube-dl #25687, #25717)
        if thumbnail_ext not in ['jpg', 'png']:

            # NB: % is supposed to be escaped with %% but this does not work
            # for input files so working around with standard substitution
            escaped_thumbnail_filename = thumbnail_filename.replace('%', '#')
            os.rename(thumbnail_filename, escaped_thumbnail_filename)
            escaped_thumbnail_jpg_filename = self.replace_extension(
                escaped_thumbnail_filename,
                'jpg',
            )

            # Run FFmpeg to convert the thumbnail(s)
            success_flag, msg = self.run_ffmpeg(
                escaped_thumbnail_filename,
                escaped_thumbnail_jpg_filename,
                ['-bsf:v', 'mjpeg2jpeg'],
            )

            if not success_flag:

                # Conversion failed; most likely because FFmpeg is not
                #   installed
                # Rename back to unescaped
                os.rename(escaped_thumbnail_filename, thumbnail_filename)

                return False

            else:

                # Conversion succeeded
                self.app_obj.remove_file(escaped_thumbnail_filename)
                thumbnail_jpg_filename = self.replace_extension(
                    thumbnail_filename,
                    'jpg',
                )

                # Rename back to unescaped for further processing
                os.rename(
                    escaped_thumbnail_jpg_filename,
                    thumbnail_jpg_filename
                )

        # Procedure complete
        return True


    def _ffmpeg_filename_argument(self, path):

        """Called by self.run_ffmpeg_multiple_files().

        Adapted from youtube-dl/youtube-dl/postprocessor/ffmpeg.py.

        Returns a filename in a format that won't confuse FFmpeg.

        Args:

            path (str): The full path to a file to be processed by FFmpeg

        Returns:

            The modified string

        """

        # Always use 'file:' because the filename may contain ':' (ffmpeg
        #   interprets that as a protocol) or can start with '-' (-- is broken
        #   in  ffmpeg, see https://ffmpeg.org/trac/ffmpeg/ticket/2127 for
        #   details)
        # Also leave '-' intact in order not to break streaming to stdout
        return 'file:' + path if path != '-' else path


    def get_executable(self):

        """Called by self.run_ffmpeg_multiple_files().

        Not adapted from youtube-dl.

        Returns the path to the FFmpeg executable, which the user may have
        specified themselves. If not, assume ffmpeg is in the system path.

        Returns:

            The path to the executable

        """

        if self.app_obj.ffmpeg_path:
            return self.app_obj.ffmpeg_path
        else:
            return 'ffmpeg'


    def is_webp(self, path):

        """Called by self.convert_webp() and utils.find_thumbnail_webp().

        Adapted from youtube-dl/youtube-dl/postprocessor/embedthumbnail.py.

        Tests whether a file is a .webp file (perhaps mislabelled as a .jpg
        file).

        Args:

            path (str): The full path to a file to be processed by FFmpeg

        """

        with open(path, 'rb') as fh:
            data = fh.read(12)

        return data[0:4] == b'RIFF' and data[8:] == b'WEBP'


    def replace_extension(self, path, ext, expected_real_ext=None):

        """Called by self.convert_webp().

        Adapted from youtube-dl/youtube-dl/utils.py.

        Given the full path to a file, replaces the extension, and returns the
        modified path.

        Args:

            path (str): The full path to a file

            ext (str): The new file extension

            expected_real_ext (str): Not used by Tartube

        Returns:

            The modified path

        """

        name, real_ext = os.path.splitext(path)

        return '{0}.{1}'.format(
            name if not expected_real_ext \
            or real_ext[1:] == expected_real_ext \
            else path,
            ext,
        )


    def run_ffmpeg(self, input_path, out_path, opt_list, test_flag=False):

        """Can be called by anything (currently called only by
        self.convert_webp() ).

        Adapted from youtube-dl/youtube-dl/postprocessor/ffmpeg.py.

        self.run_ffmpeg_multiple_files() expects a list of files. Pass on
        this function's parameters in the expected format.

        Args:

            input_path (str): Full path to a file to be processed by FFmpeg

            out_path (str): Full path to FFmpeg's output file

            opt_list (list): List of FFmpeg command line options (may be an
                empty list)

            test_flag (bool): If True, just returns the FFmpeg system command,
                rather than executing it

        Return values:

            Returns a list of two items, in the form
                (success_flag, optional_message)

        """

        return self.run_ffmpeg_multiple_files(
            [ input_path ],
            out_path,
            opt_list,
            test_flag,
        )


    def run_ffmpeg_multiple_files(self, input_path_list, out_path, opt_list, \
    test_flag=False):

        """Can be called by anything (currently called only by
        self.run_ffmpeg() ).

        Adapted from youtube-dl/youtube-dl/postprocessor/ffmpeg.py.

        Prepares the FFmpeg system command, and then executes it.

        Args:

            input_path_list (list): List of full paths to files to be
                processed by FFmpeg. At the moment, Tartube only processes one
                file at a time

            out_path (str): Full path to FFmpeg's output file

            opt_list (list): List of FFmpeg command line options (may be an
                empty list)

            test_flag (bool): If True, just returns the FFmpeg system command,
                rather than executing it

        Return values:

            Returns a list of two items, in the form
                (success_flag, optional_message)

        """

        # Get the modification time for the oldest file
        oldest_mtime = min(os.stat(path).st_mtime for path in input_path_list)

        # Prepare the system command
        files_cmd_list = []
        for path in input_path_list:
            files_cmd_list.extend(['-i', self._ffmpeg_filename_argument(path)])

        cmd_list = [self.get_executable(), '-y']
        cmd_list += ['-loglevel', 'repeat+info']
        cmd_list += (
            files_cmd_list
            + opt_list
            + [self._ffmpeg_filename_argument(out_path)]
        )

        # Return the system command only, if required
        if test_flag:
            return [ True, cmd_list ]

        # Execute the system command in a subprocess
        try:
            p = subprocess.Popen(
                cmd_list,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
            )

        except:
            # If FFmpeg is not installed on the user's system, this is the
            #   result
            return [ False, 'Could not find FFmpeg' ]

        stdout, stderr = p.communicate()
        if p.returncode != 0:
            stderr = stderr.decode(utils.get_encoding(), 'replace')
            return [ False, stderr.strip().split('\n')[-1] ]

        else:
            return [ self.try_utime(out_path, oldest_mtime, oldest_mtime), '' ]


    def run_ffmpeg_with_options(self, video_obj, source_path, cmd_list):

        """Modified version of self.run_ffmpeg(), called by
        process.ProcessManager.process_video().

        Adapted from youtube-dl/youtube-dl/postprocessor/ffmpeg.py.

        Prepares the FFmpeg system command, and then executes it.

        Args:

            video_obj (media.Video): The video object to be processed

            source_path (str): The full path to the source file

            cmd_list (list): The FFmpeg system command to use, as a list

        Return values:

            Returns a list of two items, in the form
                (success_flag, optional_message)

        """

        # Get the file's modification time
        mod_time = os.stat(source_path).st_mtime

        # Execute the system command in a subprocess
        try:
            p = subprocess.Popen(
                cmd_list,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
            )

        except:
            # If FFmpeg is not installed on the user's system, this is the
            #   result
            return [ False, 'Could not find FFmpeg' ]

        stdout, stderr = p.communicate()
        if p.returncode != 0:
            stderr = stderr.decode(utils.get_encoding(), 'replace')
            return [ False, stderr.strip().split('\n')[-1] ]

        else:
            return [ self.try_utime(source_path, mod_time, mod_time), '' ]


    def try_utime(self, path, atime, mtime):

        """Called by self.run_ffmpeg_multiple_files().

        Adapted from youtube-dl/youtube-dl/postprocessor/common.py.

        Return values:

            True on success, False on failure

        """

        try:
            os.utime(path, (atime, mtime))
            return True

        except Exception:
            return False


class FFmpegOptionsManager(object):

    """This class handles options to passed to FFmpeg, when the user wants to
    process video(s) directly (i.e. not via youtube-dl).

    Adapted from FFmpeg Command Line Wizard, by AndreKR
        (https://github.com/AndreKR/ffmpeg-command-line-wizard).

    OPTIONS (NAME TAB)

        extra_cmd_string (str): A string of extra FFmpeg options, to be applied
            to the system command (if an empty string, nothing is added)

    OPTIONS (FILE TAB)

        add_end_filename (str): A string to be added to the end of the
            filename, converting the source file to an output file with a
            different name (if an empty string, nothing is added)

        regex_match_filename (str):
        regex_apply_subst (str):
            Two strings used in a regex substituion. If the first regex matches
            the filename, the second string is used in the substitution. This
            converts the source file to an output file with a different name.
            (If the first string is an empty string, the filename is not
            changed)

        rename_both_flag (bool): If True, when a video is renamed (using any of
            the options above), the thumbnail is also renamed (but not
            vice-versa; renaming the thumbnail does not affect the video)

        change_file_ext (str): A new file extension. If not an empty string,
            then FFmpeg converts the video/audio/image file from one format to
            another. (If an empty string, the format is not changed. Ignored
            if 'output_mode' is 'gif')

        delete_original_flag (bool): If True and the source/output files have
            different names, then the source file is deleted

    OPTIONS (SETTINGS TAB)

        input_mode (str): 'video' to convert a video/audio file (the one
            downloaded by the specified media.Video object), or 'thumb' to
            convert that video's thumbnail (if it has been downloaded)

        output_mode (str): Always set to 'thumb' if 'input_mode' is set to
            'thumb'. Otherwise, the default value is 'h264'. The other possible
            values are 'gif' (in which case the video format is changed to
            .GIF), 'merge' (in which case the video is merged with au audio
            file with the same file name), 'split' (in which case the video is
            split into pieces according to timestamps) or 'slice' (In which
            case slices are removed from a video); in all four cases, the
            'change_file_ext' option is ignored)

        audio_flag (bool): If True, and if 'input_mode' is 'video', the video's
            audio is preserved when the file is converted. Ignored if
            'input_mode' is 'thumb'

        audio_bitrate (int): Value that's a multiple of 16 (minimum value is
            16)

        quality_mode (str): 'crf' for 'Manual rate factor', or 'abr' for
            'Determine from target bitrate (2-Pass)'

        rate_factor (int): Value in the range 0-51. Ignored if 'quality_mode'
            is 'abr'

        dummy_file (str): A dummy file is created during the first pass. The
            name of that file: 'output', 'dummy', '/dev/null/' for Linux,
            'NUL' for MS Windows. Ignored if 'quality_mode' is 'crf'

        patience_preset (str): Affects how long the file conversion takes, and
            also the size of the output file. Values are those used by FFmpeg
            itself: 'ultrafast', 'superfast', 'veryfast', 'faster', 'fast',
            'medium', 'slow', 'slower', 'veryslow'

        gpu_encoding (str): Optimisations for various GPUs. One of the values
            'libx264', 'libx265', 'h264_amf', 'hevc_amf', 'h264_nvenc',
            'hevc_nvenc'

        hw_accel (str): Hardware acceleration mode: 'none', 'auto', 'vdpau',
            'dxva2', 'vaapi', 'qsv'

        palette_mode (str): Ignored unless 'output_mode' is 'gif'. Values are
            'faster' or 'better'

        split_mode (str): Ignored unless 'output_mode' is 'split'. Values are
            'video' to use the video's timestamps to split the file, or
            'custom' to use a custom set of timestamps to split the file

        split_list (list): Ignored unless 'split_mode' is 'custom'. A list
            of timestamps and clip titles used to split the video into clips.
            A list in groups of three, in the form
                [start_stamp, stop_stamp, clip_title]
            If 'stop_stamp' is None, then the end of the video (or the next
                'start_stamp') is used. 'clip_title' is optional (None if not
                specified)

        slice_mode (str): Ignored unless 'output_mode' is 'slice'. Values are
            'video' to remove slices from the video using the video's own
            slice list, or 'custom' to use a custom set of slices

        slice_list (list): Ignored unless 'slice_mode' is 'custom'. A list of
            video slice data used to remove slices from a video, in the form
            described by media.Video.__init__()

    OPTIONS (OPTIMISATIONS TAB)

        seek_flag (bool): True to optimise for fast seeking (shorter keyframe
            interval, about 10% larger file)

        tuning_film_flag (bool): True if the input video is a high-quality
            movie

        tuning_animation_flag (bool): True if the input video is an animated
            movie

        tuning_grain_flag (bool): True if the input video contains film grain

        tuning_still_image_flag (bool): True if the input video is an image
            slideshow

        tuning_fast_decode_flag (bool): True to optimise for really weak CPU
            playback devices

        profile_flag (bool): True to optimise for really old devices (requires
            rate factor above 0)

        fast_start_flag (bool): True to move headers to the beginning of the
            file (so it can play while still downloading)

        tuning_zero_latency_flag (bool): True for fast encoding and low
            latency streaming

        limit_flag (bool): True to limit the bitrate, using the values
            specified by the options 'limit_mbps' and 'limit_buffer'

        limit_mbps (int): Bitrate limit in Mbit/s. Value that's a multiple of
            0.2 (minimum value is 0). Ignored if 'limit_flag' is False

        limit_buffer (int): Assume a receiving buffer (in seconds), Value
            that's a multiple of 0.2 (minimum value is 0). Ignored if
            'limit_flag' is False

    """


    # Standard class methods


    def __init__(self, uid, name):

        # IV list - other
        # ---------------
        # Unique ID for this options manager
        self.uid = uid
        # A non-unique name for this options manager
        self.name = name

        # Dictionary of FFmpeg options, set by a call to self.reset_options
        self.options_dict = {}


        # Code
        # ----

        # Initialise FFmpeg options
        self.reset_options()


    # Public class methods


    def clone_options(self, other_options_manager_obj):

        """Called by mainapp.TartubeApp.clone_ffmpeg_options() and
        .clone_ffmpeg_options_from_window().

        Clones FFmpeg options from the specified object into this object,
        completely replacing this object's FFmpeg options.

        Args:

            other_options_manager_obj (ffmpeg_tartube.FFmpegOptionsManager):
                The FFmpeg options object (usually the current one), from which
                options will be cloned

        """

        # (All values are scalars; there are no lists/dictionaries to copy)
        self.options_dict = other_options_manager_obj.options_dict.copy()


    def reset_options(self):

        """Called by self.__init__().

        Resets (or initialises) self.options_dict to its default state.
        """

        self.options_dict = {
            # NAME TAB
            'extra_cmd_string': '',
            # FILE TAB
            'add_end_filename': '',
            'regex_match_filename': '',
            'regex_apply_subst': '',
            'rename_both_flag': False,
            'change_file_ext': '',
            'delete_original_flag': False,
            # SETTINGS TAB
             # 'video', 'thumb'
            'input_mode': 'video',
            # 'h264', 'gif', 'merge', 'split', 'thumb'
            'output_mode': 'h264',
            # SETTINGS TAB ('output_mode' = 'h264')
            'audio_flag': True,
            'audio_bitrate': 128,
            # 'cfg', 'abr'
            'quality_mode': 'crf',
            'rate_factor': 23,
            # 'output', 'dummy', '/dev/null/', 'NUL'
            'dummy_file': 'output',
            # 'ultrafast', 'superfast', 'veryfast', 'faster', 'fast', 'medium',
            #   'slow', 'slower', 'veryslow'
            'patience_preset': 'medium',
            # 'libx264', 'libx265', 'h264_amf', 'hevc_amf', 'h264_nvenc',
            #   'hevc_nvenc'
            'gpu_encoding': 'libx264',
            # 'none', 'auto', 'vdpau', 'dxva2', 'vaapi', 'qsv'
            'hw_accel': 'none',
            # SETTINGS TAB ('output_mode' = 'gif')
            'palette_mode': 'faster',       # 'faster', 'better'
            # SETTINGS TAB ('output_mode' = 'split')
            'split_mode': 'video',          # 'video', 'custom'
            'split_list': [],               # list in groups of 3
            # SETTINGS TAB ('output_mode' = 'slice')
            'slice_mode': 'video',          # 'video', 'custom'
            'slice_list': [],               # list of dictionaries
            # OPTIMISATIONS TAB ('output_mode' = 'h264')
            'seek_flag': True,
            'tuning_film_flag': False,
            'tuning_animation_flag': False,
            'tuning_grain_flag': False,
            'tuning_still_image_flag': False,
            'tuning_fast_decode_flag': False,
            'profile_flag': False,
            'fast_start_flag': True,
            'tuning_zero_latency_flag': False,
            'limit_flag': False,
            'limit_mbps': 1,
            'limit_buffer': 2,
            # NOT VISIBLE IN THE EDIT WINDOW (a constant value)
            'bitrate': 0,
        }


    def get_system_cmd(self, app_obj, video_obj=None, start_point=None,
    stop_point=None, clip_title=None, clip_dir=None, edit_dict=[]):

        """Can be called by anything.

        Given the FFmpeg options specified by self.options_dict, generates the
        FFmpeg system command, returning it as a list of options.

        N.B. The 'delete_original_flag' option is not applied until the end of
        the process operation.

        Args:

            app_obj (mainapp.TartubeApp): The main application

            video_obj (media.Video or None): If specified, uses the video's
                downloaded file as the source file. Not specified when called
                from config.FFmpegOptionsEditWin, in which case a specimen
                source file is used (so that a specimen system command can be
                displayed in the edit window)

            start_point, stop_points, clip_title, clip_dir (str):
                When splitting a video, the points at which to start/stop
                (timestamps or values in seconds), the clip title, and the
                destination directory for sections (if not the same as the
                original file). Ignored if 'output_mode' is not 'split' or
                'slice'. If 'output_mode' is 'split' or 'slice', then these
                arguments are not specified when called from
                config.FFmpegOptionsEditWin, in which case specimen timestamps/
                titles are used

            edit_dict (dict): When called from the edit window, any changes
                that have been made to the FFmpeg options, but which have not
                yet been saved to this object. We take those changes into
                account when compiling the system command.

        Return values:

            Returns a tuple of three items:

                - The full path to the source file
                - The full path to the output file
                - A (python) list of options comprising the complete system
                    commmand (including the FFmpeg binary and the source/output
                    files)

        """

        opt_list = []
        tuning_list = []
        return_list = []

        # When called from the edit window (config.FFmpegOptionsEditWin), any
        #   changes in the edit window may not have been applied to
        #   self.options_dict yet
        # To produce an up-to-date system command, use a temporary copy of
        #   self.options_dict, to which the unapplied changes have been added
        options_dict = self.options_dict.copy()
        for key in edit_dict:
            options_dict[key] = edit_dict[key]

        # (Shortcuts to values retrieved several times)
        bitrate = options_dict['bitrate']
        input_mode = options_dict['input_mode']
        limit_buffer = options_dict['limit_buffer']
        limit_mbps = options_dict['limit_mbps']
        output_mode = options_dict['output_mode']
        rate_factor = options_dict['rate_factor']

        # The 'extra_cmd_string' item must be processed, and split into
        #   a list of separate items, preserving everything inside quotes as a
        #   single item (just as we do for youtube-dl download options)
        extra_cmd_string = options_dict['extra_cmd_string']
        if extra_cmd_string != '':
            extra_cmd_list = utils.parse_options(extra_cmd_string)
        else:
            extra_cmd_list = []

        # FFmpeg binary
        binary = app_obj.ffmpeg_manager_obj.get_executable()

        # Set variables describing the full path to the source video/audio and/
        #   or source thumbnail files

        # If no media.Video object was specified, then use specimen paths that
        #   can be displayed in the edit window's textview
        if video_obj is None:

            source_video_path = 'source.ext'
            source_audio_path = 'source.ext'
            source_thumb_path = 'source.jpg'

        else:

            if video_obj.dummy_flag:

                # (Special case: 'dummy' video objects (those downloaded in the
                #   Classic Mode tab) use different IVs)

                # If a specified media.Video has an unknown path, then return
                #   an empty list; there is nothing for FFmpeg to convert
                if video_obj.dummy_path is None:
                    return None

                # Check the video/audio file actually exists. If not, there is
                #   nothing for FFmpeg to convert
                source_video_path = video_obj.dummy_path
                if not os.path.exists(source_video_path) \
                and input_mode == 'video':
                    return None, None, []

            else:

                # If a specified media.Video has an unknown filename, then
                #   return an empty list; there is nothing for FFmpeg to
                #   convert
                if video_obj.file_name is None:
                    return None, None, []

                # Check the video/audio file actually exists, and is marked as
                #   downloaded. If not, there is nothing for FFmpeg to convert
                source_video_path = video_obj.get_actual_path(app_obj)
                if (
                    not os.path.exists(source_video_path) \
                    or not video_obj.dl_flag
                ) and input_mode == 'video':
                    return None, None, []

            # Find the video's thumbnail
            source_thumb_path = utils.find_thumbnail(app_obj, video_obj, True)
            if source_thumb_path is None and input_mode == 'thumb':
                # Return an empty list; there is nothing for FFmpeg to convert
                return None, None, []

            # If 'output_mode' is 'merge', then look for an audio file with the
            #   same name as the video file (but otherwise don't bother)
            source_audio_path = None
            if output_mode == 'merge':

                name, video_ext = os.path.splitext(source_video_path)
                for audio_ext in formats.AUDIO_FORMAT_LIST:

                    audio_path = os.path.abspath(os.path.join(name, audio_ext))
                    if os.path.isfile(audio_path):
                        source_audio_path = audio_path
                        break

                if source_audio_path is None:
                    # Nothing merge
                    return None, None, []

        # Break down the full path into its components, so that we can set the
        #   output file, after applying optional modifications
        if input_mode == 'video':
            output_dir, output_file = os.path.split(source_video_path)
        else:
            output_dir, output_file = os.path.split(source_thumb_path)

        output_name, output_ext = os.path.splitext(output_file)

        add_end_filename = options_dict['add_end_filename']
        if add_end_filename != '':

            # Remove trailing whitepsace
            add_end_filename = re.sub(
                r'\s+$',
                '',
                options_dict['add_end_filename'],
            )

            # Update the filename
            output_name += add_end_filename

        regex_match_filename = options_dict['regex_match_filename']
        if regex_match_filename != '':

            output_name = re.sub(
                regex_match_filename,
                options_dict['regex_apply_subst'],
                output_name,
            )

        change_file_ext = options_dict['change_file_ext'].lower()
        if change_file_ext == '':
            output_file = output_name + output_ext
        else:
            output_file = output_name + '.' + change_file_ext

        if video_obj is None:
            output_path = output_file
        else:
            output_path = os.path.abspath(
                os.path.join(output_dir, output_file),
            )

        # Special case: when called from config.FFmpegOptionsEditWin, then show
        #   a specimen system command resembling the one that will eventually
        #   be generated by FFmpegManager.run_ffmpeg_multiple_files()
        if app_obj.ffmpeg_simple_options_flag \
        and video_obj is None \
        and output_mode != 'split' \
        and output_mode != 'slice':

            return_list.append(binary)
            return_list.append('-y')
            return_list.append('-loglevel')
            return_list.append('repeat+info')
            return_list.append('-i')

            if input_mode == 'video':
                return_list.append(source_video_path)
            else:
                return_list.append(source_thumb_path)

            if extra_cmd_list:
                return_list.extend(extra_cmd_list)

            return_list.extend(
                [
                    app_obj.ffmpeg_manager_obj._ffmpeg_filename_argument(
                        output_path,
                    ),
                ],
            )

            return source_video_path, output_path, return_list

        # When the full GUI layout is visible, apply all FFmpeg options
        if input_mode == 'video':

            opt_list.append('-i')
            opt_list.append(source_video_path)

        else:

            opt_list.append('-i')
            opt_list.append(source_thumb_path)

        # H.264
        if output_mode == 'h264':

            # In the original code, this was marked:
            #   Only necessary if the output filename does not end with .mp4
            opt_list.append('-c:v')
            opt_list.append(options_dict['gpu_encoding'])

            opt_list.append('-preset')
            opt_list.append(options_dict['patience_preset'])

            if options_dict['hw_accel'] != 'none':
                opt_list.append('-hwaccel')
                opt_list.append(options_dict['hw_accel'])

            if options_dict['tuning_film_flag']:
                tuning_list.append('film')
            if options_dict['tuning_animation_flag']:
                tuning_list.append('animation')
            if options_dict['tuning_grain_flag']:
                tuning_list.append('grain')
            if options_dict['tuning_still_image_flag']:
                tuning_list.append('stillimage')
            if options_dict['tuning_fast_decode_flag']:
                tuning_list.append('fastdecode')
            if options_dict['tuning_zero_latency_flag']:
                tuning_list.append('zerolatency')

            if tuning_list:
                opt_list.append('-tune')
                opt_list.append(','.join(tuning_list))

            if options_dict['fast_start_flag']:
                opt_list.append('-movflags')
                opt_list.append('faststart')

            if input_mode == 'video' and options_dict['audio_flag']:
                opt_list.append('-c:a')
                opt_list.append('aac')
                opt_list.append('-b:a')
                opt_list.append(
                    str(options_dict['audio_bitrate']) + 'k',
                )

            if options_dict['profile_flag'] and rate_factor != 0:
                opt_list.append('-profile:v')
                opt_list.append('baseline')
                opt_list.append('-level')
                opt_list.append('3.0')

            if options_dict['limit_flag']:
                opt_list.append('-maxrate')
                opt_list.append(str(limit_mbps) + 'M')
                opt_list.append('-bufsize')
                opt_list.append((str(limit_mbps) * str(limit_buffer)) + 'M')

            if options_dict['seek_flag']:

                # In the original code, this was marked:
                #   Inserts an I-frame every 15 frames
                opt_list.append('-x264-params')
                opt_list.append('keyint=15')

            # In the original code, this was marked:
            #   Preserves the frame timestamps of VFR videos
            opt_list.append('-vsync')
            opt_list.append('2')
            opt_list.append('-enc_time_base')
            opt_list.append('-1')

            if options_dict['quality_mode'] == 'crf':

                return_list.append(binary)
                return_list.extend(opt_list)
                return_list.append('-crf')
                return_list.append(str(rate_factor))

                if extra_cmd_list:
                    return_list.extend(extra_cmd_list)

                return_list.append(output_path)

            else:
                dummy_file = options_dict['dummy_file']
                if dummy_file == 'output':
                    dummy_file = output_path

                return_list.append(binary)
                return_list.append('-y')
                return_list.extend(opt_list)
                return_list.append('-b:v')
                return_list.append(str(bitrate))
                return_list.append('-pass')
                return_list.append('1')
                return_list.append('-f')
                return_list.append('mp4')
                return_list.append(dummy_file)

                return_list.append('&&')
                return_list.append(binary)
                return_list.extend(opt_list)
                return_list.append('-b:v')
                return_list.append(str(bitrate))
                return_list.append('-pass')
                return_list.append('2')

                if extra_cmd_list:
                    return_list.extend(extra_cmd_list)

                return_list.append(output_path)

        # GIF
        elif output_mode == 'gif':

            if options_dict['palette_mode'] == 'faster':

                return_list.append(binary)
                return_list.extend(opt_list)

                if extra_cmd_list:
                    return_list.extend(extra_cmd_list)

                return_list.append(output_name + '.gif')

            else:

                return_list.append(binary)
                return_list.extend(opt_list)
                return_list.append('-vf')
                return_list.append('palettegen')
                return_list.append('palette.png')

                return_list.append('&&')
                return_list.append(binary)
                return_list.extend(opt_list)
                return_list.append('-i')
                return_list.append('palette.png')
                return_list.append('-filter_complex')
                return_list.append('"[0:v][1:v] paletteuse"')

                if extra_cmd_list:
                    return_list.extend(extra_cmd_list)

                return_list.append(output_name + '.gif')

        # Merge video/audio
        elif output_mode == 'merge':

            return_list.append(binary)
            return_list.extend(opt_list)

            return_list.append('-i')
            return_list.append(source_audio_path)
            return_list.append('-c:v')
            return_list.append('copy')
            return_list.append('-c:a')
            return_list.append('copy')

            return_list.append(output_path)

        # Split video by timestamps, or times in seconds
        elif output_mode == 'split' or output_mode == 'slice':

            return_list.append(binary)
            return_list.extend(opt_list)

            if output_mode == 'split':

                return_list.append('-ss')
                if start_point is None:
                    # (A specimen timestamp)
                    return_list.append('0:00')
                else:
                    return_list.append(str(start_point))

                # (If no timestamp is specified, the end of the video is used)
                if stop_point is not None:
                    return_list.append('-to')
                    return_list.append(str(stop_point))

            else:

                return_list.append('-ss')
                if start_point is None:
                    # (A specimen time, in seconds)
                    return_list.append('0')
                else:
                    return_list.append(str(start_point))

                # (If no timestamp is specified, the end of the video is used)
                return_list.append('-to')
                if stop_point is None:
                    # (A specimen time, in seconds)
                    return_list.append('100')
                else:
                    return_list.append(str(stop_point))

            if clip_title is None or clip_title == "":
                # (When called from config.FFmpegOptionsEditWin)
                clip_title = app_obj.split_video_generic_title

            if clip_dir is None:

                output_path = os.path.abspath(
                    os.path.join(output_dir, clip_title + output_ext),
                )

            else:

                output_path = os.path.abspath(
                    os.path.join(clip_dir, clip_title + output_ext),
                )

            return_list.append(output_path)

        # Video thumbnails
        else:

            return_list.append(binary)
            return_list.extend(opt_list)
            return_list.append(output_path)

        # All done
        if output_mode == 'thumb':
            return source_thumb_path, output_path, return_list
        else:
            return source_video_path, output_path, return_list
